;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; eyelink.lisp: lispworks library for the eyelink 1000 eyetracker
;;; Requires the eyelink api.
;;;
;;; Jelmer Borst (jpborst@ai.rug.nl)
;;; Trudy Buwalda (tabuwalda@gmail.com)
;;; Nienke van Tellingen (n.van.tellingen@student.rug.nl)
;;;
;;; Artificial Intelligence / University of Groningen
;;; 101012
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; This file contains three areas:
;;; - key-codes: contains a mapping of lisp keyboard codes to eyelink keyboard codes, for non
;;;   ascii characters.
;;; - eyelink c functions to lisp: the interface between lisp and the eyelink c code. For each 
;;;   c function that is used for the eyetracker, a lisp equivalent is defined.
;;; - lisp functions: functions for initializing the eyetracker, closing it down etc.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Typical usage:
;;; 1. before the experiment, make changes to the configuration in (initialize-eyetracker), in
;;     the settings area.
;;; 2. at the start of the experiment, call (eyelink:initialize-eyetracker (&optional filename)
;;;    This open the connection to the eyetracker, opens a edf file on the eyetracker (either
;;;    called test.edf or filename.edf), tells the eye-tracker the display properties of the 
;;;    display computer, and sends the eyetracker the configuration.
;;;    2.optional. Name the error messages log file by calling (eyelink:set-error name).
;;; 3. call (eyelink:calibrate) for each block of your experiment, for which you want to 
;;;    recalibrate the eyetracker
;;; 4. for each trial do
;;;    a) (eyelink:drift-correction) to do a drift correction in the center of the screen
;;;    b) (eyelink:start-recording 1 1 0 0) or with other settings, to start the eyetracker
;;;    c) do your stuff and send messages to the eyetracker with (eyelink:eyelink-send-message 
;;;       "message")
;;;    d) (eyelink:stop-recording)
;;; 5. call (eyelink:close-eyetracker) to close the file, copy the file to your harddisk (in the
;;;    eyetracking-data directory) and close the connection.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



(defpackage :eyelink
  (:use :cl-user :common-lisp :lispworks :mp)
  (:export 
   :initialize-eyetracker
   :close-eyetracker
   :calibrate
   :drift-correction
   :start-rec
   :stop-rec
   :send-msg
   :eyelink-time
   :set-error
   :press-eyelink-key
   ))

(in-package :eyelink)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; settings - change this section for each experiment
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defparameter *eyelink-handles-keys* nil) ;does the eyelink library handle keypresses, or your exp?
(defparameter *eyelink-handles-drift* nil) ;does the eyelink library handle drift correction display, or the experiment?
;and the rest, add later.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; key-codes
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defparameter *key-codes* (make-hash-table :test 'equalp))

(let
    ((char-list '(:LEFT :RIGHT :UP :DOWN :F1))
     (uni-list '(#0x4B00 #0x4D00 #0x4800 #0x5000 #0x3B00)))
  (if (equalp (length char-list) (length uni-list))
      (dotimes (idx (length char-list))
        (setf (gethash (nth idx char-list) *key-codes*) (nth idx uni-list)))
    (format t "Lists have unequal length.")))





;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; eyelink c functions to lisp
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;includes etc

(objc:ensure-objc-initialized :modules
'("/System/Library/Frameworks/Foundation.framework/Versions/C/Foundation"
"/Library/Frameworks/eyelink_core.framework/Versions/A/eyelink_core"
"/Library/Frameworks/eyelink_core_graphics.framework/Versions/A/eyelink_core_graphics"
"/System/Library/Frameworks/Cocoa.framework/Versions/A/Cocoa"))

;possible data types
; C lisp
; INT8 :byte
; INT16 :short
; INT32 :long
; float :float
; char (:reference :ef-mb-string :allow-null t)
; char FARTYPE * (:reference (:ef-mb-string :limit 4) :allow-null t) ??
; fartype: 32 bit = 4 byte ??


;;connection
;;;;;;;;;;;;;;;;;;;;


;;Initializes the EyeLink library, and opens a connection to the EyeLink tracker.
;Remarks:
;By setting <mode> to be 1, the connection can be simulated for debugging purposes. Only timing operations and simple tests should be done in simulation mode, and the Windows TCP/IP system must be installed. This function is intended for networks where a single tracker is connected to the network.
(fli:define-foreign-function (open-eyelink-connection
                              "open_eyelink_connection"
                              :source)
                             ((mode :short))
                             :result-type
                             :short
                             :language
                             :ansi-c)

;close_eyelink_connection()
;void close_eyelink_connection (void)
(fli:define-foreign-function (close-eyelink-connection
                              "close_eyelink_connection"
                              :source)
                              nil
                              :result-type
                              :void
                              :language
                              :ansi-c)


;;Attempts to open a link connection to the EyeLink tracker. Simple connect to single Eyelink tracker. Equivalent to eyelink_open_node(broadcast_address, 0).
(fli:define-foreign-function (eyelink-open
                              "eyelink_open"
                              :source)
                             nil
                             :result-type
                             :short
                             :language
                             :ansi-c)

;;Checks whether the connection to the tracker is alive.
(fli:define-foreign-function (eyelink-is-connected
                              "eyelink_is_connected"
                              :source)
                             nil
                             :result-type
                             :short
                             :language
                             :ansi-c)

;;Use this function to initialize the EyeLink library. This will also start the millisecond clock. No connection is attempted to the eyetracker yet. It is preferable to call open_eyelink_connection(-1) instead, as this prepares other parts of the DLL for use.
(fli:define-foreign-function (open-eyelink-system
                              "open_eyelink_system"
                              :source)
                             ((bufsize (:unsigned :short)) (options (:reference-pass :ef-mb-string :allow-null t)))
                             :result-type
                             (:unsigned :short)
                             :language
                             :ansi-c)


(fli:define-foreign-function (close-eyelink-system
                              "close_eyelink_system"
                              :source)
                             nil
                             :result-type
                             :void
                             :language
                             :ansi-c)

;set_offline_mode()
;void set_offline_mode (void)
(fli:define-foreign-function (set-offline-mode
                              "set_offline_mode"
                              :source)
                              nil
                              :result-type
                              :void
                              :language
                              :ansi-c)



;setup
;;;;;;;;;;;;;;;;;;;;

;switches to setup screen
(fli:define-foreign-function (do-tracker-setup
                              "do_tracker_setup"
                              :source)
                             nil
                             :result-type
                             :short
                             :language
                             :ansi-c)

;give commands to the eyelink
(fli:define-foreign-function (eyecmd-printf
                              "eyecmd_printf"
                              :source)
                              ((fmt :pointer)) ;(:reference-pass :ef-mb-string :allow-null t)))
                              :result-type
                              :short
                              :language
                              :ansi-c)

;set calibration target size
(fli:define-foreign-function (set-target-size
                              "set_target_size"
                              :source)
                              ((diameter (:unsigned :short)) (holesize (:unsigned :short)))
                              :result-type
                              :void
                              :language
                              :ansi-c)

;define sdl color
;(fli:define-c-typedef SDLcolor (:c-array :int 3))

;make a c struct
(fli:define-c-struct SDL_Color
  (r (:unsigned :byte))
  (g (:unsigned :byte))
  (b (:unsigned :byte))
  (unused (:unsigned :byte)))

;define this struct with a name
(fli:define-c-typedef SDLcolor (:struct SDL_Color))

;set-calibration-colors expects pointers, so we make one.
(fli:define-c-typedef SDLcolor-pointer (:pointer SDLcolor))

;set calibration colors
; void set_calibration_colors (SDL_Color *fg, SDL_Color *bg)
(fli:define-foreign-function (set-calibration-colors
                              "set_calibration_colors"
                              :source)
                              ((fg SDLcolor-pointer) (bg SDLcolor-pointer))
                              :result-type
                              :void
                              :language
                              :ansi-c)

;int16 do_drift_correct(int16 x, int16 y, int16 draw, int16 allow_setup
(fli:define-foreign-function (do-drift-correct
                              "do_drift_correct"
                              :source)
                              ((x :short) (y :short) (draw :short) (allow-setup :short))
                              :result-type
                              :short
                              :language
                              :ansi-c)



; recording
;;;;;;;;;;;;;;;;;;;;

;INT16 start_recording (INT16 file_samples, INT16 file_events, INT16 link_samples,
;INT16 link_events)
(fli:define-foreign-function (start-recording
                              "start_recording"
                              :source)
                              ((file_samples :short)(file_events :short)(link_samples :short)(link_events :short))
                              :result-type
                              :short
                              :language
                              :ansi-c)


;stop_recording()
;void stop_recording (void)
(fli:define-foreign-function (stop-recording
                              "stop_recording"
                              :source)
                              nil
                              :result-type
                              :void
                              :language
                              :ansi-c)



;file system
;;;;;;;;;;;;;

;int16 open_data_file(char *name)
(fli:define-foreign-function (open-data-file
                              "open_data_file"
                              :source)
                              ((name :pointer)) ;(:reference-pass :ef-mb-string :allow-null t)))
                              :result-type
                              :short
                              :language
                              :ansi-c)

;int16 close_data_file(void)
(fli:define-foreign-function (close-data-file
                              "close_data_file"
                              :source)
                              nil
                              :result-type
                              :short
                              :language
                              :ansi-c)

;int32 receive_data_file(char *src, char *dest, int16 dest_is_path)
(fli:define-foreign-function (receive-data-file
                              "receive_data_file"
                              :source)
                              ((src :pointer) (dest :pointer) (dest-is-path :short))
                              :result-type
                              :long
                              :language
                              :ansi-c)


; messaging
;;;;;;;;;;;;

;;;Send Message
(fli:define-foreign-function (eyelink-send-message
                              "eyelink_send_message"
                              :source)
                             ((string :pointer)) ;(:reference-pass :ef-mb-string :allow-null t)))
                            :result-type
                            :short
                            :language
                            :ansi-c)

;UINT32 eyelink_tracker_msec (void)
(fli:define-foreign-function (eyelink-tracker-time
                              "eyelink_tracker_msec"
                              :source)
                             nil
                            :result-type
                            (:unsigned :long)
                            :language
                            :ansi-c)



;graphics
;;;;;;;;;;;;;;;;;;;;

;make a c struct
(fli:define-c-struct displayinfo
  (left :long)
  (top :long)
  (right :long)
  (bottom :long)
  (width :long)
  (height :long)
  (bits :long)
  (palsize :long)
  (palrsvd :long)
  (pages :long)
  (refresh :float)
  (winnt :long))

;define this struct with a name
(fli:define-c-typedef disp (:struct displayinfo))

;however, get-display-info expects a pointer. so we need one.
(fli:define-c-typedef disp-pointer (:pointer disp))

;get-display-info
(fli:define-foreign-function 
    (get-display-information "get_display_information" :source)
    ((di disp-pointer))
  :result-type :void
  :language :ansi-c)

;create a window: call (init-expt-graphics 0 display)
;NB this function is not complete: it expect and SDL_Surface instead of an int.
;This is not necessary for letting it open its own window, which is easier anyway.
(fli:define-foreign-function (init-expt-graphics "init_expt_graphics" :source)
   ((hwnd :long) (info disp-pointer))
  :result-type 
  :short
  :language 
  :ansi-c)

;void close_expt_graphics (void)
(fli:define-foreign-function (close-expt-graphics "close_expt_graphics" :source)
    nil
    :result-type 
    :void
    :language 
    :ansi-c) 


;keyboard forwarding
;;;;;;;;;;;;;;;;;;;;;;;

;initialize getkey system
;void flush_getkey_queue(void)
(fli:define-foreign-function (flush-getkey-queue "flush_getkey_queue" :source)
    nil
    :result-type 
    :void
    :language 
    :ansi-c)

;send keys from lisp to c
(fli:define-foreign-function (eyelink-send-keybutton "eyelink_send_keybutton" :source)
    ((code (:unsigned :short)) (mods (:unsigned :short)) (state :short))
  :result-type
  :short
  :language
  :ansi-c)

;see last keys that were send
(fli:define-foreign-function (eyelink-read-keybutton "eyelink_read_keybutton" :source)
    ((mods (:pointer :long)) (state (:pointer :long)) (kcode (:pointer (:unsigned :short))) (time (:pointer (:unsigned :long))))
  :result-type
  (:unsigned :short)
  :language
  :ansi-c)

;read_getkey_queue
(fli:define-foreign-function (read-getkey-queue "read_getkey_queue" :source)
    nil
  :result-type
  (:unsigned :short)
  :language
  :ansi-c)




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; lisp functions
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;macro for safely evaluating stuff - if bus error, just continue till there's no error anymore
(defmacro safe ((type) &body body)
  `(let ((no-error nil)
        (cnt_err 0)
        (rv nil))
    (do
        ()
        (no-error)
      (handler-case
          (progn
            (setf rv (progn ,@body))
            (setf no-error t))
        (error ()
          (progn
            (incf cnt_err)
            (write-error ,type))))
      (when (> cnt_err 50)
          (print "Too many errors, bailing out...")
          (return -1)))
    rv))



;parameter that contains the dispaly info
;is set with initialize-eyetracker, so don't change the display afterwards
(defparameter *edf-file* nil) ;contains the filename of the edf-file
(defparameter *eyelink-path* (current-pathname)) ;the path of the eyelink file

(defparameter *error-name* "errors")


;print errors to error file. set this filename by calling (set-error)
(defun set-error (name)
  (setf *error-name* name))

(defun write-error (type)
  (safe
   ("write error")
   (with-open-file (out (ensure-directories-exist 
                         (merge-pathnames 
                          (make-pathname :name *error-name* :directory '(:relative "eyetracking-data") :type "txt") 
                          *eyelink-path*))
                        :direction :output :if-does-not-exist :create :if-exists :append)
     (format out "~%error in ~a" type))))


;initialize the eyetracker: call this once at the start of the experiment
;opens the connection & sends settings to the eyetracker
;filename contains filename, if not edf file will be called test.edf
(defun initialize-eyetracker (&optional (filename "test.edf"))
  (safe
   ("ini eyetracker")
  (fli:with-dynamic-foreign-objects-with-cleanups 
   ()
   (unwind-protect
       (progn
         ;first close old connection
         (safe
             ("old connection")
           (when (equalp (eyelink-is-connected) 1)
             (close-eyetracker)))
         
         ;open the connection
         (safe 
             ("new connection")
           (open-eyelink-connection 0)
           (when (equalp (eyelink-is-connected) 0)
             (format t "~%Connection to the eyelink failed.")
             (return-from initialize-eyetracker -1)))
     
         ;set offline mode
         (safe
             ("set offline mode")
           (set-offline-mode))
     
         ;open a data file ; make sure the filename is maximum 8 char long
         (open-file filename)
           
         ;tell it it was made by lisp
         (send-cmd "add_file_preamble_text 'RECORDED BY LISPWORKS 5.1' ")
     
    ;get display info
     (let 
         ((display (safe ("allocate display") (fli:allocate-foreign-object :type 'disp))))
       (safe
           ("get display info")
         (get-display-information display))
       
       (safe
           ("set disp info")
         (setf (fli:foreign-slot-value display 'right) 1023)
         (setf (fli:foreign-slot-value display 'bottom) 767)
         (setf (fli:foreign-slot-value display 'width ) 1024)
         (setf (fli:foreign-slot-value display 'height) 768))

       ;display test
       #|
       (print "-----start display------")
       (print (fli:foreign-slot-value display 'left))
       (print (fli:foreign-slot-value display 'top ))
       (print (fli:foreign-slot-value display 'right ))
       (print (fli:foreign-slot-value display 'bottom))
       (print (fli:foreign-slot-value display 'width ))
       (print (fli:foreign-slot-value display 'height))
       (print  (fli:foreign-slot-value display 'bits ))
       (print  (fli:foreign-slot-value display 'palsize))
       (print (fli:foreign-slot-value display 'palrsvd))
       (print   (fli:foreign-slot-value display 'pages ))
       (print  (fli:foreign-slot-value display 'refresh ))
       (print (fli:foreign-slot-value display 'winnt ))
       (print "-------end display-------")
       |#
     
       ;send display info to eyetracker and log it
       (safe
           ("send display to eyelink")
         (let ((top (fli:foreign-slot-value display 'top)) 
               (bottom (fli:foreign-slot-value display 'bottom)) 
               (right (fli:foreign-slot-value display 'right)) 
               (left (fli:foreign-slot-value display 'left)))
           (send-cmd (format nil "screen_pixel_coords = ~a ~a ~a ~a" top left right bottom)) ;set display size
           (send-msg (format nil "DISPLAY_COORDS ~a ~a ~a ~a" top left right bottom)))) ;log it
       
       ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
       ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
       ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
       
       
       ; settings -- this area should be changed per experiment
       ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  
       ;what are we tracking?
       (send-cmd "binocular_enabled = NO") ;only one eye
       (send-cmd "active_eye = LEFT") ;by default the left one
       (send-cmd "pupil_size_diameter = YES") ;we want pupil diameter instead of area
       
       ;calibration-type
       (send-cmd "calibration_type = HV13") ;set calibration type
       (send-cmd "enable_automatic_calibration = YES") ;automatic calibration
       (send-cmd "automatic_calibration_pacing = 1250") ;automatic calibration pacing: 1000 good for most subject, 1500 for slow ones
       
      ;calibration-size
       (safe
           ("make cal size")
         (let* ((w (fli:foreign-slot-value display 'width))
                (cal-size (round (/ w 60))) ;size of calibration target
                (focal-size (round (/ w 300))))
           (safe
               ("set target size")
             (set-target-size cal-size focal-size))))
         
       
       ;calibration-colors: should match the experiment
       (safe
           ("make cal colors")
         (let ((fg (fli:allocate-foreign-object :type 'SDLColor))
               (bg (fli:allocate-foreign-object :type 'SDLColor)))
           (setf (fli:foreign-slot-value bg 'r) 255)
           (setf (fli:foreign-slot-value bg 'g) 255)
           (setf (fli:foreign-slot-value bg 'b) 255)
           (setf (fli:foreign-slot-value fg 'r) 0)
           (setf (fli:foreign-slot-value fg 'g) 0)
           (setf (fli:foreign-slot-value fg 'b) 0)
         (safe
             ("set calibration colors")
           (set-calibration-colors fg bg))))
     
     
       ;parser configuration
       (send-cmd "select_parser_configuration 0")
     
       ;edf file contents and link data
       (send-cmd "file_event_filter = LEFT,RIGHT,FIXATION,SACCADE,BLINK,MESSAGE,BUTTON")
       (send-cmd "file_sample_data = LEFT,RIGHT,GAZE,AREA,GAZERES,STATUS")
       (send-cmd "link_event_filter = LEFT,RIGHT,FIXATION,SACCADE,BLINK,BUTTON")
       (send-cmd "link_sample_data = LEFT,RIGHT,GAZE,GAZERES,AREA,STATUS")
     
       ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
       ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
       ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
     
     ))))))



;close the link to the eyetracker, and retrieve the file
;call this once at the end of your experiment
(defun close-eyetracker ()
  (safe
   ("close eyetracker")
  (fli:with-dynamic-foreign-objects-with-cleanups 
   ()
   (unwind-protect
       (progn
         ;set offline mode
         (safe
             ("set offline mode")
           (set-offline-mode)) ;puts the eyetracker in offline mode
     
         ;give eyetracker time to finish
         (mp:wait-processing-events 1)
         
         ;close file and retrieve it
         (safe
             ("close data file")
           (close-data-file))
         
         (let* ((path 
                 (ensure-directories-exist 
                  (merge-pathnames 
                   (make-pathname :name (subseq *edf-file* 0 (position #\. *edf-file*)) :directory '(:relative "eyetracking-data") :type "edf") *eyelink-path*)))
                (name (namestring path))
                (newfilename-c (fli:allocate-foreign-object :type :ef-mb-string :initial-element name))
                (edf-file-c (fli:allocate-foreign-object :type :ef-mb-string :initial-element *edf-file*)))
           (safe
               ("receive data file")
             (receive-data-file edf-file-c newfilename-c 0)))
         
        ;close down the system
         (safe
             ("close connection")
           (close-eyelink-connection)))))))


;calibration: call this whenever you want to calibrate, usually at the start of an experiment/block
(defun calibrate ()
  (safe
   ("calibrate")
  (fli:with-dynamic-foreign-objects-with-cleanups 
   ()
   (unwind-protect
       (progn
         (if (equalp (safe ("check for connection") (eyelink-is-connected)) 1) ;check for connection
             (progn
               (let ((display (safe ("allocate display") (fli:allocate-foreign-object :type 'disp))))
                 (safe
                     ("get display info")
                   (get-display-information display))
                 (safe
                     ("set disp info")
                   (setf (fli:foreign-slot-value display 'right) 1023)
                   (setf (fli:foreign-slot-value display 'bottom) 767)
                   (setf (fli:foreign-slot-value display 'width ) 1024)
                   (setf (fli:foreign-slot-value display 'height) 768))
                 (safe
                     ("init graphics")
                      (init-expt-graphics 0 display)) ;create onscreen stuff
                 ;start key system
                 (safe
                     ("flush getkey")
                   (flush-getkey-queue))
                 (let ((liswin (if *eyelink-handles-keys* (create-listening-window) nil))) ;create a capi window to listen to keypresses
                   (safe
                       ("do tracker setup")
                     (do-tracker-setup)) ;setup the tracker, waits for return
                   (if *eyelink-handles-keys* (capi:apply-in-pane-process liswin 'capi:destroy liswin))
                   (safe
                       ("close graphics")
                     (close-expt-graphics))))) ;close the window
           (format t "~%No connection to the eyetracker.")))))))
     

;call this for every trial to correct for drift
(defun drift-correction ()
  (safe
   ("drift-correct")
   (fli:with-dynamic-foreign-objects-with-cleanups 
   ()
   (unwind-protect
       (progn
         (if (equalp (safe ("check for connection") (eyelink-is-connected)) 1) ;check for connection
             (progn
               (let ((display (safe ("allocate display") (fli:allocate-foreign-object :type 'disp))))
                 (if *eyelink-handles-drift* 
                     (progn
                       (safe
                        ("get display info")
                        (get-display-information display))
                       (safe
                        ("set disp info")
                        (setf (fli:foreign-slot-value display 'right) 1023)
                        (setf (fli:foreign-slot-value display 'bottom) 767)
                        (setf (fli:foreign-slot-value display 'width ) 1024)
                        (setf (fli:foreign-slot-value display 'height) 768))
                       (safe
                        ("init graphics")
                        (init-expt-graphics 0 display)))) ;create onscreen stuff
                  ;start key system
                 (safe
                  ("flush getkey")
                  (flush-getkey-queue))
                 (let ((liswin (if *eyelink-handles-keys* (create-listening-window) nil))
                       (success nil)) ;create a capi window to listen to keypresses
                   (safe 
                    ("do drift correct")
                    (setf success (do-drift-correct ;run drift correction
                                   512  ;x target
                                   384  ;y target
                                   (if *eyelink-handles-drift* 1 0) (if *eyelink-handles-drift*  1 0)))) ;draw, setup
                     (if *eyelink-handles-keys* (capi:apply-in-pane-process liswin 'capi:destroy liswin)) ;destroy the listening window
                     (safe
                      ("close graphics")
                      (if *eyelink-handles-drift* (close-expt-graphics)))
                     (if (equalp success 27) (calibrate)))))
               (format t "~%No connection to the eyetracker.")))))))


;let the eytracker listen to the keyboard - workaround

;function to operate the eyetracker with keyboard of display pc
(defun press-eyelink-key (pane x y gesture)
  (declare (ignore pane)
           (ignore x)
           (ignore y))
  (ignore-errors
    (let* ((data (sys:gesture-spec-data gesture)) ;get the data of the keypress: a character code or e.g. :F1
           (key (code-char data)) ;check if it's a standard charcater
           (code (if key data (gethash data *key-codes*)))) ;if it is, just send to c, otherwise look up the code in the *key-codes* hashtable
      (if (not (null code)) (safe ("send key") (eyelink-send-keybutton code 0 10)))))) ;send it to they eyetracker

;interface for listening to keypresses
(capi:define-interface listen-window ()
  ()
  (:panes
   )
  (:layouts
   (main-layout capi:pinboard-layout  ;complete window layout
                '()
                :background :white
                :input-model '(((:gesture-spec) press-eyelink-key))
                :accessor main-layout))
  (:default-initargs :title "Listening..." ;titel
   :visible-min-width 5
   :visible-min-height 5
   :x 0
   :y 0
   ))

;make a listening window
(defun create-listening-window ()
  (safe
   ("create window")
  (let ((interfaces (capi:collect-interfaces 'listen-window))) ;destroy all old st-window interfaces
    (if interfaces (mapcar #'(lambda (interface) 
                               (capi:apply-in-pane-process interface 'capi:destroy interface))
                           interfaces)))  
  (let ((listenwin (make-instance 'listen-window))) ;create a new st-window object
    (capi:display listenwin);display the window
    listenwin)))


;;message-passing with protected correct memory and safe evals
(defun send-msg (msg)
  (safe
   ("send msg out")
  (fli:with-dynamic-foreign-objects-with-cleanups 
   ((c-msg (:ef-mb-string) :initial-element msg))
   (unwind-protect
       (safe
           ("send msg")
         (eyelink-send-message c-msg))))))

(defun send-cmd (cmd)
  (safe
   ("send cmd out")
  (fli:with-dynamic-foreign-objects-with-cleanups 
   ((c-cmd (:ef-mb-string) :initial-element cmd))
   (unwind-protect
       (safe
           ("send cmd")
         (eyecmd-printf c-cmd))))))

(defun open-file (filename)
  (safe
   ("open file out")
  (let* ((f filename)
         (edf (if (equalp (subseq f (- (length f) 4) (length f)) ".edf") f (format nil "~a~a" f ".edf")))
         (edf8 (if (> (length edf) 12) (format nil "~a.edf" (subseq f 0 8)) edf)))
    (setf *edf-file* edf8)
    (fli:with-dynamic-foreign-objects-with-cleanups 
     ((c-file (:ef-mb-string) :initial-element edf8))
     (unwind-protect
         (safe
             ("open file")
           (open-data-file c-file)))))))


(defun start-rec (a b c d)
  (safe
      ("start rec")
    (start-recording a b c d)))

(defun stop-rec ()
  (safe
      ("stop rec")
    (stop-recording)))

(defun eyelink-time ()
  (safe
      ("eyelink time")
    (eyelink-tracker-time)))
